package net.codinux.banking.client.fints4k

import net.codinux.banking.client.model.*
import net.codinux.banking.client.model.AccountTransaction
import net.codinux.banking.client.model.Amount
import net.codinux.banking.client.model.tan.*
import net.codinux.banking.client.model.options.GetAccountDataOptions
import net.codinux.banking.client.model.response.*
import net.codinux.banking.client.model.tan.ActionRequiringTan
import net.codinux.banking.client.model.tan.TanChallenge
import net.codinux.banking.client.model.tan.TanImage
import net.codinux.banking.client.model.tan.TanMethod
import net.codinux.banking.client.model.tan.TanMethodType
import net.dankito.banking.client.model.BankAccountIdentifierImpl
import net.dankito.banking.client.model.parameter.GetAccountDataParameter
import net.dankito.banking.client.model.parameter.RetrieveTransactions
import net.dankito.banking.client.model.response.ErrorCode
import net.codinux.banking.fints.mapper.FinTsModelMapper
import net.codinux.banking.fints.model.*
import net.codinux.banking.fints.messages.datenelemente.implementierte.tan.TanMedium
import net.codinux.banking.fints.messages.datenelemente.implementierte.tan.MobilePhoneTanMedium
import net.codinux.banking.fints.messages.datenelemente.implementierte.tan.TanGeneratorTanMedium
import net.codinux.banking.fints.messages.datenelemente.implementierte.tan.TanMediumStatus
import net.dankito.banking.banklistcreator.prettifier.BankingGroupMapper
import kotlin.io.encoding.Base64
import kotlin.io.encoding.ExperimentalEncodingApi

open class FinTs4kMapper {

    protected val fintsModelMapper = FinTsModelMapper()

    protected val bankingGroupMapper = BankingGroupMapper()


    open fun mapToGetAccountDataParameter(credentials: AccountCredentials, options: GetAccountDataOptions) = GetAccountDataParameter(
        credentials.bankCode, credentials.loginName, credentials.password,
        options.accounts.map { BankAccountIdentifierImpl(it.identifier, it.subAccountNumber, it.iban) },
        options.retrieveBalance,
        RetrieveTransactions.valueOf(options.retrieveTransactions.name), options.retrieveTransactionsFrom, options.retrieveTransactionsTo,
        preferredTanMethods = options.preferredTanMethods?.map { mapTanMethodType(it) },
        abortIfTanIsRequired = options.abortIfTanIsRequired
    )

    protected open fun mapTanMethodType(type: TanMethodType): net.codinux.banking.fints.model.TanMethodType =
        net.codinux.banking.fints.model.TanMethodType.valueOf(type.name)


    open fun map(response: net.dankito.banking.client.model.response.GetAccountDataResponse): Response<GetAccountDataResponse> {
        return if (response.successful && response.customerAccount != null) {
            Response.success(GetAccountDataResponse(mapUser(response.customerAccount!!)))
        } else {
            mapError(response)
        }
    }


    open fun mapToUserAccountViewInfo(bank: BankData): UserAccountViewInfo = UserAccountViewInfo(
        bank.bankCode, bank.customerId, bank.bankName, getBankingGroup(bank.bankName, bank.bic)
    )

    open fun mapToBankAccountViewInfo(account: AccountData): BankAccountViewInfo = BankAccountViewInfo(
        account.accountIdentifier, account.subAccountAttribute,
        mapAccountType(fintsModelMapper.map(account.accountType)),
        account.iban, account.productName
    )


    protected open fun mapUser(user: net.dankito.banking.client.model.CustomerAccount) = UserAccount(
        user.bankCode, user.loginName, user.password,
        user.bankName, user.bic, user.customerName, user.userId,
        user.accounts.map { mapAccount(it) },

        user.selectedTanMethod?.securityFunction?.code, user.tanMethods.map { mapTanMethod(it) },
        user.selectedTanMedium?.mediumName, user.tanMedia.map { mapTanMedium(it) },

        getBankingGroup(user.bankName, user.bic)
    )

    protected open fun getBankingGroup(bankName: String, bic: String): BankingGroup? =
        bankingGroupMapper.getBankingGroup(bankName, bic)


    protected open fun mapAccount(account: net.dankito.banking.client.model.BankAccount) = BankAccount(
        account.identifier, account.accountHolderName, mapAccountType(account.type), account.iban, account.subAccountNumber,
        account.productName, account.currency, account.accountLimit, account.isAccountTypeSupportedByApplication,
        mapFeatures(account),
        mapAmount(account.balance), account.retrievedTransactionsFrom, account.retrievedTransactionsTo,
        // TODO: map haveAllTransactionsBeenRetrieved
        countDaysForWhichTransactionsAreKept = account.countDaysForWhichTransactionsAreKept,
        bookedTransactions = account.bookedTransactions.map { mapTransaction(it) }.toMutableList()
    )

    protected open fun mapAccountType(type: net.dankito.banking.client.model.BankAccountType): BankAccountType =
        BankAccountType.valueOf(type.name)

    protected open fun mapFeatures(account: net.dankito.banking.client.model.BankAccount): Set<BankAccountFeatures> = buildSet {
        if (account.supportsRetrievingBalance) {
            add(BankAccountFeatures.RetrieveBalance)
        }
        if (account.supportsRetrievingTransactions) {
            add(BankAccountFeatures.RetrieveTransactions)
        }
        if (account.supportsTransferringMoney) {
            add(BankAccountFeatures.TransferMoney)
        }
        if (account.supportsInstantPayment) {
            add(BankAccountFeatures.InstantPayment)
        }
    }


    protected open fun mapTransaction(transaction: net.dankito.banking.client.model.AccountTransaction): AccountTransaction = AccountTransaction(
        mapAmount(transaction.amount), transaction.amount.currency.code, transaction.unparsedReference,
        transaction.bookingDate, transaction.valueDate,
        transaction.otherPartyName, transaction.otherPartyBankCode, transaction.otherPartyAccountId,
        transaction.bookingText, null,
        transaction.statementNumber, transaction.sequenceNumber,
        mapNullableAmount(transaction.openingBalance), mapNullableAmount(transaction.closingBalance),

        transaction.endToEndReference, transaction.customerReference, transaction.mandateReference,
        transaction.creditorIdentifier, transaction.originatorsIdentificationCode,
        transaction.compensationAmount, transaction.originalAmount,
        transaction.sepaReference,
        transaction.deviantOriginator, transaction.deviantRecipient,
        transaction.referenceWithNoSpecialType, transaction.primaNotaNumber, transaction.textKeySupplement,

        transaction.currencyType, transaction.bookingKey,
        transaction.referenceForTheAccountOwner, transaction.referenceOfTheAccountServicingInstitution,
        transaction.supplementaryDetails,

        transaction.transactionReferenceNumber, transaction.relatedReferenceNumber
    )

    protected open fun mapNullableAmount(amount: Money?) = amount?.let { mapAmount(it) }

    protected open fun mapAmount(amount: Money) = Amount.fromString(amount.amount.string.replace(',', '.'))


    open fun mapTanChallenge(challenge: net.codinux.banking.fints.model.TanChallenge): TanChallenge {
        val type = mapTanChallengeType(challenge)
        val action = mapActionRequiringTan(challenge.forAction)

        val tanMethods = challenge.bank.tanMethodsAvailableForUser.map { mapTanMethod(it) }
        val selectedTanMethodId = challenge.tanMethod.securityFunction.code

        val tanMedia = challenge.bank.tanMedia.map { mapTanMedium(it) }
        val selectedTanMediumName = challenge.bank.selectedTanMedium?.mediumName

        val user = mapToUserAccountViewInfo(challenge.bank)
        val account = challenge.account?.let { mapToBankAccountViewInfo(it) }

        val tanImage = if (challenge is ImageTanChallenge) mapTanImage(challenge.image) else null
        val flickerCode = if (challenge is FlickerCodeTanChallenge) mapFlickerCode(challenge.flickerCode) else null

        return TanChallenge(type, action, challenge.messageToShowToUser, selectedTanMethodId, tanMethods, selectedTanMediumName, tanMedia, tanImage, flickerCode, user, account)
    }

    protected open fun mapTanChallengeType(challenge: net.codinux.banking.fints.model.TanChallenge): TanChallengeType = when {
        challenge is ImageTanChallenge -> TanChallengeType.Image
        challenge is FlickerCodeTanChallenge -> TanChallengeType.Flickercode
        else -> TanChallengeType.EnterTan
    }

    protected open fun mapActionRequiringTan(action: net.codinux.banking.fints.model.ActionRequiringTan): ActionRequiringTan =
        ActionRequiringTan.valueOf(action.name)

    protected open fun mapTanMethod(method: net.codinux.banking.fints.model.TanMethod): TanMethod = TanMethod(
        method.displayName, mapTanMethodType(method.type), method.securityFunction.code, method.maxTanInputLength, mapAllowedTanFormat(method.allowedTanFormat)
    )

    protected open fun mapTanMethodType(type: net.codinux.banking.fints.model.TanMethodType): TanMethodType =
        TanMethodType.valueOf(type.name)

    protected open fun mapAllowedTanFormat(allowedTanFormat: net.codinux.banking.fints.messages.datenelemente.implementierte.tan.AllowedTanFormat?): AllowedTanFormat =
        allowedTanFormat?.let { AllowedTanFormat.valueOf(it.name) } ?: AllowedTanFormat.Alphanumeric

    protected open fun mapTanImage(image: net.codinux.banking.fints.tan.TanImage): TanImage =
        TanImage(image.mimeType, mapToBase64(image.imageBytes), mapException(image.decodingError))

    @OptIn(ExperimentalEncodingApi::class)
    protected open fun mapToBase64(bytes: ByteArray): String {
        return Base64.Default.encode(bytes)
    }

    protected open fun mapTanMedium(tanMedium: TanMedium) = net.codinux.banking.client.model.tan.TanMedium(
        mapTanMediumType(tanMedium), tanMedium.mediumName, mapTanMediumStatus(tanMedium.status),
        (tanMedium as? TanGeneratorTanMedium)?.let { mapTanGeneratorTanMedium(it) },
        (tanMedium as? MobilePhoneTanMedium)?.let { mapMobilePhoneTanMedium(it) }
    )

    protected open fun mapTanMediumStatus(status: TanMediumStatus): net.codinux.banking.client.model.tan.TanMediumStatus = when (status) {
        TanMediumStatus.Aktiv -> net.codinux.banking.client.model.tan.TanMediumStatus.Used
        TanMediumStatus.Verfuegbar -> net.codinux.banking.client.model.tan.TanMediumStatus.Available
        TanMediumStatus.AktivFolgekarte -> net.codinux.banking.client.model.tan.TanMediumStatus.ActiveFollowUpCard
        TanMediumStatus.VerfuegbarFolgekarte -> net.codinux.banking.client.model.tan.TanMediumStatus.AvailableFollowUpCard
    }

    protected open fun mapMobilePhoneTanMedium(tanMedium: MobilePhoneTanMedium) = net.codinux.banking.client.model.tan.MobilePhoneTanMedium(
        tanMedium.phoneNumber, tanMedium.concealedPhoneNumber
    )

    protected open fun mapTanGeneratorTanMedium(tanMedium: TanGeneratorTanMedium) = net.codinux.banking.client.model.tan.TanGeneratorTanMedium(
        tanMedium.cardNumber, tanMedium.cardSequenceNumber, tanMedium.cardType,
        tanMedium.validFrom, tanMedium.validTo
    )

    protected open fun mapTanMediumType(tanMedium: TanMedium): TanMediumType = when {
        tanMedium is MobilePhoneTanMedium -> TanMediumType.MobilePhone
        tanMedium is TanGeneratorTanMedium -> TanMediumType.TanGenerator
        else -> TanMediumType.Generic
    }

    protected open fun mapFlickerCode(flickerCode: net.codinux.banking.fints.tan.FlickerCode): FlickerCode =
        FlickerCode(flickerCode.challengeHHD_UC, flickerCode.parsedDataSet, mapException(flickerCode.decodingError))


    protected open fun <T> mapError(response: net.dankito.banking.client.model.response.GetAccountDataResponse): Response<T> {
        return if (response.error != null) {
            Response.error(ErrorType.valueOf(response.error!!.name), if (response.error == ErrorCode.BankReturnedError) null else response.errorMessage,
                if (response.error == ErrorCode.BankReturnedError && response.errorMessage !== null) listOf(response.errorMessage!!) else emptyList())
        } else {
            Response.error(ErrorType.UnknownError, response.errorMessage)
        }
    }

    protected open fun mapException(exception: Exception?): String? =
        exception?.stackTraceToString()

}